library list_scaffold;

import 'dart:async';

import 'package:flutter/material.dart' hide SearchDelegate, showSearch;
import 'package:flutter/material.dart' as DefaultSearchDelegate
    show SearchDelegate, showSearch;

class ListScaffoldWidgets {
  const ListScaffoldWidgets({
    this.error = const Center(child: Text('Oops, something went wrong')),
    this.loading = const Center(child: CircularProgressIndicator()),
    this.empty = const Center(child: Text('Empty List')),
    this.noResults = const Center(child: Text('No Results')),
  });

  final Widget error, loading, empty, noResults;
}

class ListScaffold<T> extends StatefulWidget {
  ListScaffold({
    Key key,
    this.list,
    this.itemBuilder,
    this.searchFilter,
    this.searchDelegate,
    this.widgets = const ListScaffoldWidgets(),
  }) : super(key: key);

  final dynamic list;
  final Widget Function(BuildContext context, T item) itemBuilder;
  final bool Function(T item, String query) searchFilter;
  final DefaultSearchDelegate.SearchDelegate searchDelegate;
  final ListScaffoldWidgets widgets;

  @override
  ListScaffoldState<T> createState() => ListScaffoldState<T>();
}

class ListScaffoldState<T> extends State<ListScaffold<T>> {
  List<T> _searchList;

  Future<T> showSearch() async => await DefaultSearchDelegate.showSearch<T>(
        context: context,
        delegate: widget.searchDelegate ??
            _SearchDelegate<T>(
              list: widget.list is List<T> ? widget.list : _searchList ?? [],
              itemBuilder: widget.itemBuilder,
              searchFilter: widget.searchFilter,
              widgets: widget.widgets,
            ),
      );

  @override
  Widget build(BuildContext context) => widget.list is Future<List<T>>
      ? FutureBuilder<List<T>>(
          future: widget.list,
          builder: (BuildContext context, AsyncSnapshot<List<T>> snapshot) =>
              _build(
            isLoading: snapshot.connectionState == ConnectionState.waiting,
            hasError: snapshot.hasError,
            // Assign result to searchList to be accessable for search delegate
            list: _searchList ??= snapshot.data,
          ),
        )
      : (widget.list is List<T>
          // Is loading for normal list is when it's null
          ? _build(isLoading: widget.list == null, list: widget.list)
          : ErrorWidget(
              'type ${widget.list.runtimeType} is not subtype of List<$T> or Future<List<$T>>'));

  // Build All
  Widget _build({bool isLoading, bool hasError = false, List<T> list}) =>
      hasError
          ? _buildError()
          : (isLoading
              ? _buildLoading()
              : list.isEmpty ? _buildEmpty() : _buildList(list));

  // Build Loading
  Widget _buildLoading() => widget.widgets.loading;

  // Build Loading
  Widget _buildError() => widget.widgets.error;

  // Build Empty
  Widget _buildEmpty() => widget.searchFilter == null
      ? widget.widgets.noResults
      : widget.widgets.empty;

  // Build ListView
  Widget _buildList(List<T> list) => ListView.builder(
        itemCount: list.length,
        itemBuilder: (BuildContext context, int index) => widget.itemBuilder(
          context,
          list[index],
        ),
      );
}

class _SearchDelegate<T> extends DefaultSearchDelegate.SearchDelegate<T> {
  _SearchDelegate({
    this.list,
    this.itemBuilder,
    this.searchFilter,
    this.widgets = const ListScaffoldWidgets(),
  });

  final List<T> list;
  final bool Function(T item, String query) searchFilter;
  final Function(BuildContext, T) itemBuilder;
  final ListScaffoldWidgets widgets;

  @override
  Widget buildLeading(BuildContext context) => IconButton(
        tooltip: MaterialLocalizations.of(context).backButtonTooltip,
        icon: Icon(Icons.arrow_back),
        onPressed: () => close(context, null),
      );

  @override
  Widget buildResults(BuildContext context) => buildSuggestions(context);

  @override
  Widget buildSuggestions(BuildContext context) => ListScaffold<T>(
        list: list.where(_filterSuggestions).toList(),
        itemBuilder: (BuildContext context, item) => InkWell(
          child: itemBuilder(context, item),
          onTap: () => close(context, item),
        ),
        widgets: widgets,
      );

  // Filter suggestions if search filter is null set default filter
  bool _filterSuggestions(T item) => searchFilter != null
      ? searchFilter(item, query)
      : item.toString().startsWith(query.toString());

  @override
  // Build back clear button if query is empty will close else will clear
  List<Widget> buildActions(BuildContext context) => <Widget>[
        IconButton(
          tooltip: query.isEmpty
              ? MaterialLocalizations.of(context).closeButtonTooltip
              : 'Clear',
          icon: const Icon(Icons.clear),
          onPressed: () {
            if (query.isEmpty) {
              close(context, null);
            } else {
              query = '';
              showSuggestions(context);
            }
          },
        ),
      ];
}
